import os
import re
import time
import json
import md5
import pdb
import random
import psycopg2
import traceback
import numpy as np

from functools import wraps
from collections import *
from datetime import datetime
from sqlalchemy import *
from sqlalchemy.pool import NullPool
from flask import Flask, request, render_template, g, redirect, Response
from flask.ext.compress import Compress


from summary import Summary
from util import *
from db import *

tmpl_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'templates')
print tmpl_dir
app = Flask(__name__, template_folder=tmpl_dir)
Compress(app)


def returns_json(f):
  @wraps(f)
  def json_returner(*args, **kwargs):
    r = f(*args, **kwargs)
    if not isinstance(r, basestring):
      r = json.dumps(r, cls=SummaryEncoder)
    return Response(r, content_type='application/json')
  return json_returner

def cache_result(key, value):
  engine = db_connect('cache')
  db = engine.connect()
  q = "insert into requests values(%s, %s)"
  db.execute(q, key, value)
  db.close()
  engine.dispose()

@app.before_request
def before_request():
  try:
    dbname = None
    if 'db' in request.form:
      dbname = request.form['db']
    elif 'db' in request.args:
      dbname = request.args['db']

    g.dbname = dbname
    g.engine = None
    g.db = None

    if dbname:
      g.engine = db_connect(dbname)
      g.db = g.engine.connect()
  except:
    traceback.print_exc()
    g.engine = None
    g.db = None

@app.teardown_request
def teardown_request(exception):
  try:
    if hasattr(g, 'db'):
      g.db.close()
  except Exception as e:
    pass

  try:
    if hasattr(g, 'engine'):
      g.engine.dispose()
  except Exception as e:
    pass


@app.route('/', methods=["POST", "GET"])
def index():
  try:
    import scorpion
    enable_scorpion = 1
    title = 'DBWipes + Scorpion!'
  except:
    enable_scorpion = 0
    title = 'DBWipes'

  context = {
    'enableScorpion': enable_scorpion,
    'js': 'summary',
    'study': 0,
    'title': title,
    'debug': True
  }
  return render_template("index_base.html", **context)




@app.route('/drspott/', methods=["POST", "GET"])
def drspott():
  context = {
    'enableScorpion': 1,
    'js': 'summarydrspott',
    'study': 0,
    'title': 'DBWipes + Scorpion!'
  }
  return render_template("index_base.html", **context)


@app.route('/hidden/', methods=["POST", "GET"])
def hidden():
  context = {
    'enableScorpion': 1,
    'js': 'summary',
    'title': 'DBWipes + Scorpion!',
    'study': 0,
    'debug': True
  }
  return render_template("index_base.html", **context)

@app.route('/study/name/', methods=["POST", "GET"])
def study_name():
  return render_template("study/name.html")

@app.route('/study/', methods=["POST", "GET"])
@app.route('/study/dir/', methods=["POST", "GET"])
def study_dir():
  return render_template("study/dir.html", **{
    'taskids': enumerate([4,5, 8, 10])
  })

@app.route('/study/<int:idx>/', methods=["POST", "GET"])
def index_idx(idx):
  templates = []
  templates.append("study/index%d.html" % idx)
  templates.append("index_base.html")

  js = 'study/summary%d' % idx
  title = "DBWipes Tutorial"
  subtitle = ""
  enable_scorpion = 1

  if idx in [0,1,2,4]:
    enable_scorpion = 0

  if idx == 2:
    title = 'DBWipes Verification Test'

  if idx == 3:
    title = "DBWipes + Scorpion Tutorial"

  if idx >= 4:
    title = "DBWipes User Study"
    subtitle = "without Scorpion"

  # hard1 sum
  if idx == 4:
    enable_scorpion = 0

  if idx == 5:
    subtitle = "with Scorpion"
    js = 'study/summary4'
    enable_scorpion = 1
    templates[0] = "study/index4.html"

  # intel
  if idx == 6:
    subtitle = "without Scorpion"
    js = 'study/summary6'
    enable_scorpion = 0

  if idx == 7:
    subtitle = "with Scorpion"
    js = 'study/summary6'
    enable_scorpion = 1
    templates[0] = "study/index6.html"

  # hard1 avg
  if idx == 8:
    enable_scorpion = 0
    js = 'study/summary8'
    templates[0] = "study/index8.html"

  if idx == 9:
    subtitle = "with Scorpion"
    js = 'study/summary8'
    enable_scorpion = 1
    templates[0] = "study/index8.html"


  # hard2 sum
  if idx == 10:
    js = 'study/summary10'
    enable_scorpion = 0
    templates[0] = "study/index10.html"

  if idx == 11:
    subtitle = "with Scorpion"
    js = 'study/summary10'
    enable_scorpion = 1
    templates[0] = "study/index10.html"

  # hard2 avg
  if idx == 12:
    js = 'study/summary12'
    enable_scorpion = 0
    templates[0] = "study/index12.html"

  if idx == 13:
    subtitle = "with Scorpion"
    js = 'study/summary12'
    enable_scorpion = 1
    templates[0] = "study/index12.html"



  context = {
    'enableScorpion': enable_scorpion,
    'idx': idx,
    'js': js,
    'study': 1,
    'title': title,
    'subtitle': subtitle,
    'debug': False
  }
  print context

  return render_template(templates, **context)

@app.route('/tasks/get/', methods=["POST", "GET"])
@returns_json
def task_get():
  name = request.form['name']
  if not name:
    return { 'status': False }

  try:
    db = db_connect("tasks")
    try:
      q = """create table tasks(
        name varchar, 
        tstamp timestamp default current_timestamp,
        tasks text
      )"""
      db.execute(q)
    except:
      pass


    q = """select * from tasks where name = %s"""
    rows = db.execute(q, name).fetchall()
    if rows:
      tasks = json.loads(rows[0][2])
      return {
        'status': True,
        'tasks': tasks
      }
    else:
      alltasks = [4, 8, 10, 12]
      options = np.random.choice(alltasks, 3, replace=False)
      tasks = []
      for task in options:
        enable_scorpion = random.random() > .5
        if enable_scorpion:
          task += 1
        tasks.append(task)
      q = """insert into tasks values(%s, default, %s)"""
      db.execute(q, (name, json.dumps(tasks)))
      return {
        'status': True,
        'tasks': tasks
      }
  except Exception as e:
    print e
    pass
  finally:
    if db:
      db.dispose()

 

@app.route('/tasks/submit/', methods=["POST", "GET"])
@returns_json
def task_submit():
  print request.form
  name = request.form['name']
  taskid = request.form['taskid']
  data = request.form['data']
  db = db_connect("tasks")

  try:
    q = """create table responses(
      name varchar, 
      tstamp timestamp default current_timestamp,
      taskid varchar, 
      data text
    )"""
    db.execute(q)
  except:
    pass

  q = "insert into responses values(%s, default, %s, %s)"
  db.execute(q, (name, taskid, data))

  db.dispose()
  return {'status': True}
 
@app.route('/api/databases/', methods=['POST', 'GET'])
@returns_json
def dbs():
  q = "SELECT datname FROM pg_database where datistemplate = false;"
  dbnames = [str(row[0]) for row in g.db.execute(q).fetchall()]
  return {'databases': dbnames}



@app.route('/api/tables/', methods=['POST', 'GET'])
@returns_json
def tables():
  cur = g.db.cursor()
  res = cur.execute("SELECT tablename from pg_tables WHERE schemaname = 'public'")
  tables = [str(row[0]) for row in res.fetchall()]
  return {'tables': tables}



def get_schema(db_or_name, table):
  summary = Summary(db_or_name, table)
  try:
    cols_and_types = summary.get_columns_and_types()
    schema = dict(cols_and_types)
    return schema
  except Exception as e:
    traceback.print_exc()
  finally:
    summary.close()
  return {}



@app.route('/api/schema/', methods=['POST', 'GET'])
@returns_json
def schema():
  table = request.args.get('table')
  
  if not table:
    return {}

  ret = {}
  ret['schema'] = get_schema(g.db, table)
  return ret


@app.route('/api/requestid/', methods=['POST', 'GET'])
@returns_json
def requestid():
  try:
    from scorpion.util import Status
    status = Status()
    requestid = status.reqid
    status.close()
    return {'requestid': requestid}
  except Exception as e:
    return {'error': str(e)}



@app.route('/api/status/', methods=['POST', 'GET'])
@returns_json
def api_status():
  try:
    from scorpion.util import Status
    rid = int(request.args.get('requestid'))

    status = Status(rid)
    ret = status.latest_status()
    label_rules = status.get_rules()
    status.close()

    partial_rules = []
    for label, rules in label_rules:
      partial_rules.extend(rules)
    rules_hash = hash(str(partial_rules))

    return {
      'status': ret,
      'results': partial_rules,
      'hash': rules_hash
    }
  except Exception as e:
    return {
      'status': str(e),
      'results': []
    }


@app.route('/api/tuples/', methods=['POST', 'GET'])
@returns_json
def api_tuples():
  ret = { }
  jsonstr = request.args.get('json')
  if not jsonstr:
    print "query: no json string.  giving up"
    return ret

  args = json.loads(jsonstr)
  dbname = args.get('db')
  table = args.get('table')
  where = args.get('where', []) or []

  where, params = where_to_sql(where)
  if where:
    where = 'AND %s' % where
  print where
  print params

  query = """WITH XXXX as (select count(*) from %s WHERE 1 = 1 %s)
  SELECT * FROM %s 
  WHERE random() <= 50.0 / (select * from XXXX) %s 
  LIMIT 50"""
  query = query % (table, where, table, where)
  try:
    conn = g.db
    cur = conn.execute(query, [params+params])
    rows = cur.fetchall()
    cur.close()

    data = [dict(zip(cur.keys(), vals)) for vals in rows]
    ret['data'] = data
    ret['schema'] = get_schema(g.db, table)

  except Exception as e:
    traceback.print_exc()
    ret = {}
    raise

  print "%d points returned" % len(ret.get('data', []))
  return ret



@app.route('/api/query/', methods=['POST', 'GET'])
@returns_json
def api_query():
  ret = { }
  jsonstr = request.args.get('json')
  if not jsonstr:
    print "query: no json string.  giving up"
    return ret

  args = json.loads(jsonstr)
  dbname = args.get('db')
  table = args.get('table')

  o, params = create_sql_obj(g.db, args)
  o.limit = 10000;
  query = str(o)
  print query
  print params

  if not dbname or not table or not query:
    print "query: no db/table/query.  giving up"
    return ret

  try:
    conn = g.db
    cur = conn.execute(query, [params])
    rows = cur.fetchall()
    cur.close()

    data = [dict(zip(cur.keys(), vals)) for vals in rows]
    ret['data'] = data
    ret['schema'] = get_schema(g.db, table)

  except Exception as e:
    traceback.print_exc()
    ret = {}

  print "%d points returned" % len(ret.get('data', []))
  return ret



@app.route('/api/column_distribution/', methods=['POST', 'GET'])
@returns_json
def column_distribution():
  dbname = request.args.get('db', 'intel')
  tablename = request.args.get('table', 'readings')
  where = request.args.get('where', '')
  col = request.args.get('col')
  try:
    nbuckets = int(request.args.get('nbuckets', 100))
  except Exception as e:
    print e
    nbuckets = 100


  summary = Summary(g.db, tablename, nbuckets=nbuckets, where=where)
  try:
    typ = summary.get_type(col)
    stats = summary.get_col_stats(col, typ)
  except Exception as e:
    traceback.print_exc()
  finally:
    summary.close()

  data = {
    'col': col,
    'type': typ,
    'stats': stats
  }
  context = { "data": data }
  return context



@app.route('/api/column_distributions/', methods=['POST', 'GET'])
@returns_json
def column_distributions():
  dbname = request.args.get('db', 'intel')
  tablename = request.args.get('table', 'readings')
  where = request.args.get('where', '')
  try:
    nbuckets = int(request.args.get('nbuckets', 100))
  except Exception as e:
    print e
    nbuckets = 100

  #  #from monetdb import sql as msql
  #  #db = msql.connect(user='monetdb', password='monetdb', database=dbname)

  summary = Summary(g.db, tablename, nbuckets=nbuckets, where=where)
  print 'where: %s' % where
  try:
    stats = summary()
  except Exception as e:
    traceback.print_exc()
  finally:
    summary.close()


  data = []
  for col, typ, col_stats in stats:
    data.append({
      'col': col, 
      'type': typ, 
      'stats': col_stats
    })

  context = { "data": data }
  return context



@app.route('/api/scorpion/', methods=['POST', 'GET'])
@returns_json
def scorpion():
  try:
    import scorpionutil
  except:
    print >>sys.stderr, "Could not load scorpionutil.  Maybe scorpion has not been installed?"
    traceback.print_exc()
    return {}

  try:
    data =  json.loads(str(request.form['json']))
    fake = request.form.get('fake', False)
    requestid = request.form.get('requestid')
    if not fake or fake == 'false':
      results = scorpionutil.scorpion_run(g.db, data, requestid)
      return results
  except:
    return {}

  ret = {}
  results = [
    {
      'score': 0.2,
      'c_range': [0, 1],
      'count': 100,
      'clauses': [
        {'col': 'sensor', 'type': 'str', 'vals': map(str, [18])}
      ],
      'alt_rules': [
        [ {'col': 'humidity', 'type': 'num', 'vals': [0, 1.4]}]
      ]
    },
    {
      'score': 0.2,
      'c_range': [0, 1],
      'count': 100,
      'clauses': [
        {'col': 'voltage', 'type': 'num', 'vals': [0, 2.15]},
        {'col': 'sensor', 'type': 'str', 'vals': ['18']}
      ],
      'alt_rules': [
        [ {'col': 'humidity', 'type': 'num', 'vals': [0, 1.4]},
          {'col': 'humidity', 'type': 'num', 'vals': [0, 1.4]} ],
        [ {'col': 'humidity', 'type': 'num', 'vals': [0, 1.4]},
          {'col': 'humidity', 'type': 'num', 'vals': [0, 1.4]} ]
      ]
    }
  ]

  top_k = [
    {
      'c': 0,
      'score': 0.2,
      'c_range': [0, 0],
      'count': 100,
      'clauses': [
        {'col': 'sensor', 'type': 'str', 'vals': map(str, [18])}
      ]
    },
    {
      'c': 0,
      'score': 0.2,
      'c_range': [0, 0],
      'count': 100,
      'clauses': [
        {'col': 'voltage', 'type': 'num', 'vals': [0, 2.15]},
        {'col': 'sensor', 'type': 'str', 'vals': ['18']}
      ]
    },
    {
      'c': 0.5,
      'score': 0.2,
      'c_range': [0.5, .5],
      'count': 100,
      'clauses': [
        {'col': 'sensor', 'type': 'str', 'vals': map(str, [18, 15])}
      ]
    },
    {
      'c': 0.5,
      'score': 0.2,
      'c_range': [0.5, .5],
      'count': 100,
      'clauses': [
        {'col': 'voltage', 'type': 'num', 'vals': [-5, 2.5]},
        {'col': 'sensor', 'type': 'str', 'vals': ['18', '15']}
      ]
    },
    {
      'c': 1.0,
      'score': 0.2,
      'c_range': [1.0, 1.0],
      'count': 100,
      'clauses': [
        {'col': 'sensor', 'type': 'str', 'vals': map(str, [18, 30, 35])}
      ]
    },
    {
      'c': 1.0,
      'score': 0.2,
      'c_range': [1.0, 1.0],
      'count': 100,
      'clauses': [
        {'col': 'humidity', 'type': 'num', 'vals': [-100, 40]},
        {'col': 'sensor', 'type': 'str', 'vals': ['18', '19']}
      ]
    }

  ]

  from scorpion.util import Status
  status = Status(requestid)
  status.update_rules('label', results)
  status.close()

  time.sleep(1)

  ret['results'] = results
  ret['top_k_results'] = top_k
  return ret





import os
import re
import time
import json
import md5
import pdb
import psycopg2
import traceback
import numpy.random as nprand

from collections import *
from datetime import datetime

from scorpionsql.aggerror import AggErr
from scorpion.arch import SharedObj, extract_agg_vals
from scorpion.parallel import parallel_debug
from scorpion.util import Status

from util import *
from db import db_type


def scorpion_run(db, requestdata, requestid):
  """
  badsel:  { alias: { x:, y:, xalias:, yalias:, } }
  goodsel: { alias: { x:, y:, xalias:, yalias:, } }
  """
  context = {}

  try:
    qjson = requestdata.get('query', {})
    tablename = qjson['table']
    parsed, params = create_sql_obj(db, qjson)
    print "parsed SQL"
    print parsed
  except Exception as e:
    traceback.print_exc()
    context["error"] = str(e)
    return context



  try:    
    badsel = requestdata.get('badselection', {})
    goodsel = requestdata.get('goodselection', {})
    errtypes = requestdata.get('errtypes', {})
    erreqs = requestdata.get('erreqs', {})
    ignore_attrs = requestdata.get('ignore_cols', [])
    dbname = qjson['db']
    x = qjson['x']
    ys = qjson['ys']

    obj = SharedObj(db, dbname=dbname, parsed=parsed, params=params)
    obj.dbname = dbname
    obj.C = 0.2
    obj.ignore_attrs = map(str, ignore_attrs )

    # fix aliases in select
    for nonagg in obj.parsed.select.nonaggs:
      nonagg.alias = x['alias']
    for agg in obj.parsed.select.aggs:
      y = [y for y in ys if y['expr'] == agg.expr][0]
      agg.alias = y['alias']


    xtype = db_type(db, tablename, x['col'])

    errors = []
    for agg in obj.parsed.select.aggregates:
      alias = agg.shortname
      if alias not in badsel:
        continue

      badpts = badsel.get(alias, [])
      badkeys = map(lambda pt: pt['x'], badpts)
      badkeys = extract_agg_vals(badkeys, xtype)
      goodpts = goodsel.get(alias, [])
      goodkeys = map(lambda pt: pt['x'], goodpts)
      goodkeys = extract_agg_vals(goodkeys, xtype)
      errtype = errtypes[alias]
      print "errtype", errtype
      erreq = []
      if errtype == 1:
        erreq = erreqs[alias]
        print "erreq", erreq


      # XXX: for better end-to-end performance
      #      sample the badkeys and goodkeys so there are ~20
      if len(badkeys) > 20:
        idxs = range(len(badkeys))
        newidxs = nprand.choice(idxs, 20, False)
        badkeys = map(badkeys.__getitem__, newidxs)
        if erreq:
          erreq = map(erreq.__getitem__, newidxs)
      if len(goodkeys) > 20:
        idxs = range(len(goodkeys))
        newidxs = nprand.choice(idxs, 20, False)
        goodkeys = map(goodkeys.__getitem__, newidxs)


      err = AggErr(agg, badkeys, 20, errtype, {'erreq': erreq})
      obj.goodkeys[alias] = goodkeys
      errors.append(err)

    obj.errors = errors

    obj.status = Status(requestid)
    print "status requid = ", requestid

    start = time.time()
    parallel_debug(
      obj,
      parallel=True,
      nstds=0,
      errperc=0.001,
      epsilon=0.008,
      msethreshold=0.15,
      tau=[0.001, 0.05],
      c=obj.c,
      complexity_multiplier=4.5,
      l=0.7,
      c_range=[0.1, 1.],
      max_wait=20,
      use_cache=False,
      granularity=20,
      ignore_attrs=obj.ignore_attrs,
      DEBUG=False
    )
    cost = time.time() - start
    print "end to end took %.4f" % cost


    obj.update_status('serializing results')
    context['results'] = encode_best_rules(obj)
    context['top_k_results'] = encode_top_k(obj)

    obj.update_status('done!')

  except Exception as e:
    traceback.print_exc()
    context['error'] = str(e)
  finally:
    try:
      obj.status.close()
    except:
      pass


  return context

def encode_best_rules(obj):
  """
  Given the resulting rule clusters, package them into renderable
  JSON objects
  """
  from scorpion.learners.cn2sd.rule import rule_to_json
  results = []
  for yalias, rules in obj.rules.items():
    for rule in rules:
      result = rule_to_json(rule, yalias=yalias)
      results.append(result)
  return results

def encode_top_k(obj):
  """
  extract the top_k per c_value and encode as a dictionary:
  {
    c_val: [json_rules in descending influence order]
  }
  """
  from scorpion.learners.cn2sd.rule import rule_to_json
  results = [] #defaultdict(lambda: defaultdict(list))
  for yalias, top_k_rules in obj.top_k_rules.items():
    for c_val, rules in top_k_rules.iteritems():
      for rule in rules:
        rule.c_range = [c_val, c_val]
        result = rule_to_json(rule, yalias=yalias)
        result['c'] = round(c_val, 3)
        results.append(result)
        #results[yalias][c_val].append(result)
  return results





#!/usr/bin/env python2.7

try:
  activate_this = './bin/activate_this.py'
  execfile(activate_this, dict(__file__=activate_this))
except:
  pass


import click
import sys
import psycopg2
from dbwipes.server import app

DEC2FLOAT = psycopg2.extensions.new_type(
  psycopg2.extensions.DECIMAL.values,
  'DEC2FLOAT',
  lambda value, curs: float(value) if value is not None else None)
print "registering type"
psycopg2.extensions.register_type(DEC2FLOAT)


@click.command()
@click.option('--debug', is_flag=True)
@click.option('--threaded', is_flag=True)
@click.argument('HOST', default='localhost')
@click.argument('PORT', default=8111, type=int)
def run(debug, threaded, host, port):
  HOST, PORT = host, port
  print "running on %s:%d" % (HOST, PORT)

  app.run(host=HOST, port=PORT, debug=debug, threaded=threaded)


  try:
    from tornado .wsgi import WSGIContainer
    from tornado.httpserver import HTTPServer
    from tornado.ioloop import IOLoop

    print "running tornado server"
    http_server = HTTPServer(WSGIContainer(app))
    http_server.listen(PORT, address=HOST)
    IOLoop.instance().start()
  except Exception as e:
    print e
    try:
      
      from gevent.wsgi import WSGIServer
      print "running gevent server"
      http_server = WSGIServer((HOST, PORT), app)
      http_server.serve_forever()
    except:
      app.debug = True
      print "running flask server"
      app.run(host=HOST, port=PORT, debug=debug)


if __name__ == '__main__':
  run()



#!/usr/bin/env python2.7

try:
  activate_this = './bin/activate_this.py'
  execfile(activate_this, dict(__file__=activate_this))
except:
  pass

import click
import sys
from sqlalchemy import *
from dbwipes.summary import Summary


@click.command()
@click.option('--reset', is_flag=True)
@click.argument('dbname')
@click.argument('tables', nargs=-1)
def script(reset, dbname, tables):
  def get_tables(db):
    q = "select tablename from pg_tables where schemaname = 'public'"
    return [str(table) for (table,) in db.execute(q).fetchall()]

  if not tables:
    engine = create_engine("postgresql://localhost/%s" % dbname)
    db = engine.connect()
    tables = get_tables(db)
    db.close()
    engine.dispose()

  for table in tables:
    summary = Summary(dbname, table, nbuckets=500)
    if reset:
      summary.reset_cache()
    summary()
    summary.close()


if __name__ == '__main__':
  script()






import sys
import pdb
import psycopg2

from sqlalchemy import create_engine
from sqlalchemy.pool import NullPool
from datetime import datetime, date, timedelta
from datetime import time as dttime
from collections import deque, defaultdict
from dateutil.parser import parse as dateparse

from settings import *


def db_connect(dbname, engine='pg'):
  try:
    conn = "%s/%s" % (DBPREFIX, dbname)
    db = create_engine(conn, poolclass=NullPool)
  except:
    sys.stderr.write( "couldn't connect\n")
    sys.exit()
  return db


def db_type(db, table, col):
  q = """SELECT pg_type.typname FROM pg_attribute, pg_class, pg_type where 
    relname = %%s and pg_class.oid = pg_attribute.attrelid and attname = '%s' and
    pg_type.oid = atttypid"""
  try:
    row = db.execute(q % col, table).fetchone()
    return row[0]
  except Exception as e:
    print e
    return None


def db_schema(db, table):
  """
  only works for postgres

  @return dictionary of column name -> python type object
  """
  typedict = [('int', int), ('double', float),
              ('timestamp', datetime),
              ('date', date), ('time', dttime),
              ('text', str), ('char', str)]
  ret = {}
  q = '''select column_name, data_type
          from information_schema.columns
          where table_name = %s;'''
  #q = """select c.name, c.type from sys.columns as c, sys.tables as t where table_id = t.id and t.name = %s;"""
  # and data_type != 'date' and position('time' in data_type) =
  # 0 and column_name != 'humidity'
  for row in query(db, q, (table,)):
    name, dtype = tuple( row[:2] )
    name = str(name)
    for tn, tt in typedict:
      if tn in dtype:
        ret[name] = tt
        break
    if name not in ret:
      msg = "can't find type of %s\t%s"
      raise RuntimeError(msg % (name, dtype))
  return ret


def close(db):
  if db != None:
    db.close()




#
# this file has no deps on Scorpion
#
import os
import re
import time
import json
import md5
import pdb
import psycopg2
import traceback

from collections import *
from datetime import datetime


from scorpionsql.errfunc import *
from scorpionsql.sql import *


# JSON Encoder
class SummaryEncoder(json.JSONEncoder):
  def default(self, o):
    if isinstance(o, float):
      if o == float('inf'):
        return 1e100
      elif o == float('-inf'):
        return -1e100

    if hasattr(o, 'isoformat'):
      s =  o.isoformat()
      if not s.endswith("Z"):
        s += 'Z'
      return s
    return super(SummaryEncoder, self).default(o)




def where_to_sql(where_json, negate=False):
  is_type = lambda s, types: any([t in s for t in types])
  l = []
  args = []
  for clause_json in where_json:
    if 'sql' in clause_json:
      l.append(clause_json['sql'])
      continue

    ctype = clause_json['type']
    col = clause_json['col']
    vals = clause_json['vals']


    if not vals: continue

    if is_type(ctype, ['num', 'int', 'float', 'double', 'date', 'time']):
      q = "%%s <= %s and %s <= %%s" % (col, col)
      args.extend(vals)
    else:
      tmp = []
      vals = list(vals)
      if None in vals:
        tmp.append("(%s is null)" % col)

      realvals = list(filter(lambda v: v is not None, vals))
      if len(realvals) == 1:
        tmp.append("(%s = %%s)" % col)
        args.append(realvals[0])
      elif len(realvals):
        tmp.append("(%s in %%s)" % col)
        args.append(tuple(list(realvals)))
      q = ' or '.join(tmp)

    l.append(q)

  q = ' and '.join(filter(bool, l))
  if negate and q:
    q = "not(%s)" % q
  return q, args


__agg2f__ = {
  'avg' : AvgErrFunc,
  'std' : StdErrFunc,
  'stddev' : StdErrFunc,
  'stddev_samp': StdErrFunc,
  'stddev_pop': StdErrFunc,
  'min' : MinErrFunc,
  'max' : MaxErrFunc,
  'sum' : SumErrFunc,
  'corr' : CorrErrFunc,
  'count' : CountErrFunc,
  'abs' : AbsErrFunc
}


def parse_agg(s):
  """
  parse an aggregation SELECT clause e.g., avg(temp) as foo
  into dictionary of function name, column, and alias components
  """
  print s
  p = re.compile('(?P<func>\w+)\(\s*(?P<col>[\w\,\s]+)\s*\)\s*(as\s+(?P<alias>\w+))?')
  d = p.match(s).groupdict()
  klass = __agg2f__[d['func'].strip()]
  expr = str(d['col'])
  cols = [col.strip() for col in expr.split(',')]
  varlist = [Var(col) for col in cols]
  print klass
  print cols
  print varlist
  func = klass(varlist)
  return {
    'fname': d['func'],
    'func': func,
    'cols': cols,
    'alias': d.get('alias', '') or d['func']
  }

def expr_from_nonagg(s):
  """
  remove alias component of a nonaggregation SELECT clause
  """
  if ' as ' in s: 
    return ' as '.join(s.split(' as ')[:-1])
  return s


def create_sql_obj(db, qjson):
  x = qjson['x']
  ys = qjson['ys']
  #sql = qjson['query']
  dbname = qjson['db']
  table = qjson['table']
  negate = qjson.get('negate', False)
  where_json = qjson.get('where', []) or []
  basewheres_json = qjson.get('basewheres', []) or []

  where, args = where_to_sql(where_json, negate)
  basewheres, baseargs = where_to_sql(basewheres_json, False)
  where = ' and '.join(filter(bool, [where, basewheres]))
  args.extend(baseargs)
  
  select = Select()
  nonagg = SelectExpr(x['alias'], [x['col']], x['expr'], x['col'])
  select.append(nonagg)
  for y in ys:
    d = parse_agg(y['expr'])
    agg = SelectAgg(y['alias'], d['func'], d['cols'], y['expr'], d['cols'][0])
    select.append(agg)

  parsed = Query(
    db, 
    select, 
    [table], 
    [where],
    [x['expr']], 
    [expr_from_nonagg(x['expr'])]
  )

  return parsed, args





#!/usr/bin/env python2.7
#
# Generate synthetic test data for user study
#
# shell commands:
#
#   python gendata.py --mode hard1  > simpledata.csv
#   importmydata.py simple test simpledata.csv 
#   summarizedb.py test simple
#
#


try:
  activate_this = './bin/activate_this.py'
  execfile(activate_this, dict(__file__=activate_this))
except:
  pass


import random
import click
import numpy as np
from collections import *
random.seed(0)
np.random.seed(0)


@click.command()
@click.argument('mode')
def gen_data(mode):
  # numeric attributes
  #  date, amount, age, state
  # discrete attributes
  #  
  states = ["AL",  "AK", "AZ", "AR", "CA", "CO", "CT", 
  "DE", "DC", "FL", "GA", "HI", "ID", "IL", "IN", "IA", 
  "KS", "KY", "LA", "ME", "MD", "MA", "MI", "MN", "MS", 
  "MO", "MT", "NE", "NV", "OH", "OK", "OR", "PA", "TN", 
  "TX", "UT", "VT", "VA", "WA", "WI", "WY"]

  ages = ['<18', '18-30', '30-50', '>60']

  state2ages = defaultdict(list)
  for state in states:
    sages = np.array([random.randint(1, 40) for age in ages]).astype(float)
    sages /= sages.sum()
    state2ages[state] = sages

  state2gender = {state: random.random() for state in states}


  print "day,state,age,gender,amt"
  for date in range(10):

    if mode == 'simple':
      states = ["AR", "CA", "CO", "IL",  "PA", "TN", "VT", "WA", "WI" ]

    # 5k sales per date
    for sidx, state in enumerate(states):

      
      #
      # Simple Version
      #
      # CA Males have more sales 
      # PA Females have higher sales
      # 
      if mode == 'simple':
        numcenter = 50
        numstd = 5
        amtcenter = 100
        amtstd = 5

        if state == 'CA':
          numcenter *= 0.88 * date
          amtcenter = 70


        state2gender['PA'] = 0.60
        state2gender['CA'] = 0.9

        nsales = max(0, int(random.gauss(numcenter, numstd)))
        agelist = np.random.choice(ages, nsales, p=state2ages[state])
        for i, age in enumerate(agelist):
          damtcenter = damtstd = 0
          male = random.random() < state2gender[state]
          gender = (male and 'Male' or 'Female')

          if state == 'PA' and gender == 'Female' and date >= 3:
            damtcenter += 35 * date
            damtstd = -amtstd + 3

          amt = max(0, random.gauss(amtcenter + damtcenter, amtstd+damtstd))
          #if state in ['CA', 'PA', 'OH']:
          print "%d,%s,%s,%s,%.2f" % (date, state, age, gender, amt)


      #
      # Hard version 1
      #
      # average of (40 + idx*2) +- 5 sales ~$10 each
      # CA has higher sales
      # MI has _more_ sales
      # FL have higher than average number and rate of sales 

      elif mode == 'hard1':
        numcenter = 40 + sidx*2
        numstd = 5
        amtcenter = 80 - sidx/2.
        amtstd = 5


        if date > 4:
          if state in ['CA', 'MI']:
            numcenter *= max(1, 1.15 * (date-4))

          if state == 'MI':
            amtcenter += 30
          if state == 'CA':
            amtcenter += 30
            amtstd += 5
          if state == 'FL':
            amtcenter *= 3

        nsales = max(0, int(random.gauss(numcenter, numstd)))
        agelist = np.random.choice(ages, nsales, p=state2ages[state])
        for i, age in enumerate(agelist):
          damtcenter = damtstd = 0
          if age == '>60':
            damtcenter += 50
          if age == '<18':
            damtstd += 5
            damtcenter += 20

          male = random.random() < state2gender[state]
          gender = (male and 'M' or 'F')

          amt = max(0, random.gauss(amtcenter + damtcenter, amtstd+damtstd))

          if state == 'FL':
            male = random.random() < 0.6
            gender = (male and 'M' or 'F')
            if gender == 'F':
              damtcenter = -amtcenter-50
            amt = random.gauss(amtcenter + damtcenter, amtstd+damtstd)

          print "%d,%s,%s,%s,%.2f" % (date, state, age, gender, amt)


      #
      # Hard version 2
      #
      # average of (40 + idx*2) +- 5 sales ~$10 each
      # MA >60 has higher sales
      # WA <18 has _more_ sales

      elif mode == 'hard2':
        numcenter = 20
        numstd = 3
        amtcenter = 50
        amtstd = 2


        if date >= 3 and date <= 6:
          if state in ['MA']:
            amtcenter *= max(1, 1.15 * (6 - abs(4-date)))

          #if state == 'MA':
            #amtcenter += 30
          if state == 'WA':
            numcenter += 20
            amtcenter += 60

        nsales = max(0, int(random.gauss(numcenter, numstd)))
        perc = np.array(list(state2ages[state]))
        if state == 'MA':
          perc[3] += .7
          perc = perc / perc.sum()
        if state == 'WA':
          perc[0] += .7
          perc = perc / perc.sum()


        agelist = np.random.choice(ages, nsales, p=perc)
        for i, age in enumerate(agelist):
          damtcenter = damtstd = 0
          male = random.random() < state2gender[state]
          gender = (male and 'M' or 'F')
          amt = max(0, random.gauss(amtcenter + damtcenter, amtstd+damtstd))

          print "%d,%s,%s,%s,%.2f" % (date, state, age, gender, amt)


if __name__ == '__main__':
  gen_data()



#!/usr/bin/env python2.7
try:
    from setuptools import setup, find_packages
except ImportError:
    import ez_setup
    ez_setup.use_setuptools()
from setuptools import setup, find_packages
import dbwipes

setup(name="dbwipes",
      version=dbwipes.__version__,
      description="dbwipes a database",
      license="MIT",
      author="Eugene Wu",
      author_email="eugenewu@mit.edu",
      url="http://github.com/sirrice/summary",
      packages = find_packages(),
      include_package_data = True,      
      package_data = {
        'static': 'dbwipes/static',
        'templates': 'dbwipes/templates'
      },
      package_dir = {'dbwipes' : 'dbwipes'},
      scripts = [
        'bin/dbwipesserver.py',
        'bin/summarizedb.py',
        'bin/gen_dbwipes_test_data.py',
        'bin/setupdbwipes.sh'
      ],
      #zip_safe = False,
      install_requires = [
        'flask', 'psycopg2', 'sqlalchemy', 
        'flask-compress', 'scorpionsql', 'click'
      ],
      keywords= "")



from functools import wraps
import bsddb
import json
import pdb
from sqlalchemy import create_engine, engine

from db import *



def get_cache():
  eng = db_connect('cache')
  db = eng.connect()

  try:
    db.execute("create table cache(key varchar, val text)")
  except:
    pass

  return eng, db

def make_cache(f):
  @wraps(f)
  def _f(self, *args, **kwargs):
    try:
      key = str(map(str, (f.__name__, self.engine, self.tablename, self.where, self.nbuckets, map(str, args))))
      #print key
      vals = self._cache.execute('select val from cache where key = %s', key).fetchall()
      if len(vals):
        return json.loads(vals[0][0])
    except Exception as e:
      print e
      pdb.set_trace()

    with self._cache.begin() as txn:
      res = f(self, *args, **kwargs)
      if key:
        self._cache.execute('insert into cache values(%s, %s)', key, json.dumps(res, default=json_handler))
      return res
    return None
  return _f

def json_handler(o):
  if hasattr(o, 'isoformat'):
    return o.isoformat()


class Summary(object):

  def __init__(self, dbname, tablename, nbuckets=50, where=''):
    self.dbtype = 'pg'
    if 'monetdb' in str(dbname):
      self.engine = None
      self.db = dbname
      self.dbtype = 'monet'
    elif isinstance(dbname, engine.base.Engine):
      self.engine = dbname
      self.db = dbname.connect()
    elif isinstance(dbname, engine.base.Connection):
      self.engine = dbname.engine
      self.db = dbname
    else:
      self.engine = db_connect(dbname)
      self.db = self.engine.connect()

    self.tablename = tablename
    self.nbuckets = nbuckets
    self.where = ''
    where = where.strip()
    if where:
      self.where = 'WHERE %s' % where

    self._engine, self._cache = get_cache()

    self.nrows = self.get_num_rows()
    self.col_types = self.get_columns_and_types()

  def __call__(self):
    stats = []
    for col, typ in self.col_types:
      #print "stats for: %s\t%s" % (col, typ)
      col_stats = self.get_col_stats(col, typ)
      if col_stats is None:
        #print "\tgot None"
        continue
      #print "\tgot %d" % (len(col_stats))
      stats.append((col, typ, col_stats))
    return stats


  def close(self):
    try:
      self.db.close()
    except Exception as e:
      print e
      pass

    try:
      self.engine.dispose()
    except Exception as e:
      print e
      pass

    try:
      self._cache.close()
    except Exception as e:
      print e
      pass

    try:
      self._engine.dispose()
    except Exception as e:
      print e
      pass

    #print self._cache.closed, self.db.closed

  def reset_cache(self):
    q = """delete from cache where key like '%%%%%s%%%%%s%%%%'""" % (str(self.engine), self.tablename)
    self._cache.execute(q)

  def query(self, q, *args):
    """
    Summaries using other engines only need to override this method
    """

    if self.dbtype == 'pg':
      print q
      return self.db.execute(q, *args).fetchall()
    else:
      cur = self.db.cursor()
      try:
        print q
        print args
        if args:
          cur.execute(q, args)
        else:
          cur.execute(q)
        ret = cur.fetchall()
        return ret
      except:
        self.db.rollback()
        raise
      finally:
        cur.close()


  @make_cache
  def get_num_rows(self):
    q = "SELECT count(*) from %s" % self.tablename
    return self.query(q)[0][0]

  @make_cache
  def get_distinct_count(self, col):
    q = "SELECT count(distinct %s) FROM %s %s" % (col, self.tablename, self.where)
    return self.query(q)[0][0]

  @make_cache
  def get_column_counts(self, cols):
    q = """SELECT %s FROM %s"""
    select = ["count(distinct %s)" % col for col in cols]
    select = ", ".join(select)
    q = q % (select, self.tablename)
    counts = tuple(self.query(q)[0])
    return dict(zip(cols, counts))



  @make_cache
  def get_columns_and_types(self):
    if self.dbtype == 'pg':
      q = """
      SELECT attname, pg_type.typname 
      FROM pg_class, pg_attribute, pg_type
      WHERE relname = %s and 
            pg_attribute.attrelid = pg_class.oid and 
            pg_type.oid = atttypid and
            attnum > 0 and 
            attisdropped = false;
      """
    else:
      q = """
      SELECT columns.name , columns.type
      FROM columns, tables 
      WHERE tables.name = %s and 
            tables.id = columns.table_id;
      """
    ret = []
    for (col, typ) in self.query(q, self.tablename):
      ret.append((str(col), str(typ)))
    return ret


  @make_cache
  def get_columns(self):
    """
    engine specific way to get table columns
    """
    if self.dbtype == 'pg':
      q = "select attname from pg_class, pg_attribute where relname = %s and attrelid = pg_class.oid and attnum > 0 and attisdropped = false;"
    else:
      q = "select columns.name from columns, tables where tables.name = %s and tables.id = columns.table_id;"
    ret = []
    for (attr,) in self.query(q, self.tablename):
      ret.append(str(attr))
    return ret


  @make_cache
  def get_type(self, col_name):
    if self.dbtype == 'pg':
      q = """SELECT pg_type.typname FROM pg_attribute, pg_class, pg_type where 
        relname = %s and pg_class.oid = pg_attribute.attrelid and attname = %s and
        pg_type.oid = atttypid"""
    else:
      q = """SELECT columns.type from columns, tables 
      WHERE tables.name = %s and tables.id = columns.table_id and columns.name = %s;
      """
    try:
      rows = self.query(q, self.tablename, col_name)
      row = rows[0]
      return row[0]
    except Exception as e:
      import traceback
      traceback.print_exc()
      return None


  def get_col_groupby(self, col_name, col_type):
    if col_type == None:
      return None

    groupby = None

    if 'time' == col_type:
      groupby = self.get_time_stats(col_name)

    if 'date' in col_type or 'timestamp' in col_type:
      groupby = self.get_date_stats(col_name)

    return groupby


  @make_cache
  def get_col_stats(self, col_name, col_type=None):
    if col_type is None:
      col_type = self.get_type(col_name)

    if col_type.startswith('_'):
      return None

    numerics = ['int', 'float', 'double', 'numeric']
    is_numeric = any([s in col_type for s in numerics])
    if self.dbtype == 'pg' and is_numeric:
      stats = self.get_numeric_stats(col_name)
      return stats
    if any([s in col_type for s in ['char', 'text']]):
      return self.get_char_stats(col_name)

    groupby = self.get_col_groupby(col_name, col_type)
    if groupby:
      stats = self.get_group_stats(col_name, groupby)
      return stats
    return None



  def get_group_stats(self, col_name, groupby):
    q = """select %s as GRP, min(%s), max(%s), count(*) 
    from %s  %s group by GRP 
    order by GRP limit %d"""
    q = q % (groupby, col_name, col_name, self.tablename, self.where, self.nbuckets)
    rows = [{ 'val': x, 'count': count, 'range':[minv, maxv]} for (x, minv, maxv, count) in self.query(q)]
    return rows

  def get_numeric_stats(self, c):
    ndistinct = self.get_distinct_count(c)
    if ndistinct == 0:
      return []
    if ndistinct == 1:
      if self.where:
        q = "SELECT %s from %s %s AND %s is not null"
        args = (c, self.tablename, self.where, c)
      else:
        q = "SELECT %s from %s WHERE %s is not null"
        args = (c, self.tablename, c)
      val = self.query(q % args)[0][0]
      return [{'val': val, 'count': self.nrows, 'range': [val, val]}]
    
    q = """
    with bound as (
      SELECT min(%s) as min, max(%s) as max, avg(%s) as avg, stddev(%s) as std FROM %s %s
    )
    SELECT width_bucket(%s::numeric, (avg-2.5*std), (avg+2.5*std), %d) as bucket,
           min(%s) as min,
           max(%s) as max,
           count(*) as count
    FROM %s, bound
    %s
    GROUP BY bucket
    """
    q = q % (c, c, c, c, self.tablename, self.where, c, self.nbuckets, c, c, self.tablename, self.where)


    q = """
    with TMP as (
      SELECT 2.5 * stddev(%s) / %d as block FROM %s %s
    )
    SELECT (%s/block)::int*block as bucket, 
           min(%s) as min, 
           max(%s) as max, 
           count(*) as count
    FROM %s,  TMP
    %s
    GROUP BY bucket
    ORDER BY bucket
    """
    q = q % (c, self.nbuckets, self.tablename, self.where, c, c, c, self.tablename, self.where)


    stats = []
    for (val, minv, maxv, count) in self.query(q):
      if val is None:
        stats.append({
          'val': None,
          'count': count,
          'range': [minv, maxv]
        })
      else:
        stats.append({
          'val': (maxv+minv)/2.,
          'count': count,
          'range': [minv, maxv]
        })

    return stats


  def get_char_stats(self, col_name):
    q = """select %s as GRP, min(%s), max(%s), count(*)
    FROM %s 
    %s
    GROUP BY GRP
    ORDER BY count(*) desc
    LIMIT %d"""
    q = q % (col_name, col_name, col_name, self.tablename, self.where, self.nbuckets)
    rows = [{ 'val': x, 'count': count, 'range':[minv, maxv]} for (x, minv, maxv, count) in self.query(q)]
    return rows


    
    groupby = col_name
    return groupby
    return self.get_group_stats(col_name, groupby)

  def get_time_stats(self, col_name):
    if self.dbtype == 'pg':
      return "date_trunc('hour', %s)::time" % col_name
    return """cast((%s 
              - cast(extract(second from %s) as interval second) 
              - cast(extract(minute from %s) as interval minute))
            as time)""" % (col_name, col_name, col_name)

  def get_date_stats(self, col_name):
    if self.dbtype == "pg":
      q = "select max(%s)::date, min(%s)::date, EXTRACT(EPOCH FROM (max(%s::timestamp) - min(%s::timestamp)))/60 as minutes from %s"
    else:
      q = """select cast(max(%s) as date), cast(min(%s) as date), 
        cast(max(cast(%s as timestamp)) - min(cast(%s as timestamp)) as bigint)/1000/60 as minutes
        from %s"""
    q = q % (col_name, col_name, col_name, col_name,  self.tablename)
    (maxv, minv, nminutes) = self.query(q)[0]
    if maxv is None or minv is None or nminutes is None:
      return None


    ndays = nminutes / 60 / 24

    if self.dbtype == 'pg':
      var = "%s::timestamp" % col_name
    else:
      var = "cast(%s as timestamp)" % col_name

    if self.dbtype == 'pg':
      if ndays == 0:
        groupby = "date_trunc('hour', %s)" % var
      elif ndays <= 30:
        groupby = "date_trunc('day', %s)" % var
      elif ndays <= 50 * 7:
        groupby = "date_trunc('week', %s)" % var
      elif ndays <= 365 * 12:
        groupby = "date_trunc('month', %s)" % var
      else:
        groupby = "date_trunc('year', %s)" % var
    else:
      if ndays == 0:
        groupby = """(%s 
          - cast(extract(second from %s) as interval second) 
          - cast(extract(minute from %s) as interval minute))
        """ % (var, var, var)
      elif ndays <= 30:
        groupby = "cast(cast(%s as date) as timestamp)" % var
      elif ndays <= 365 * 12:
        groupby = "cast(%s as date) - cast(extract(day from %s) as interval day)" % (var, var)
      else:
        groupby = """(cast(%s as date) 
        - cast(extract(day from %s) as interval day)
        - cast(extract(month from %s) as interval month)
        """ % (var, var, var)
    return groupby



